﻿using System;
using System.IO;
using System.Net;

namespace NaGet
{
	/// <summary>
	/// 環境変数などを取り扱うクラス
	/// </summary>
	public sealed class Env
	{		
		/// <summary>
		/// 呼び出し禁止
		/// </summary>
		private Env()
		{
		}
		
		#region pref.xml関連
		
		/// <summary>
		/// 設定を格納
		/// </summary>
		private static NaGetLibPref pref;
		
		/// <summary>
		/// 設定（のクローンオブジェクト）を返す。
		/// </summary>
		public static NaGetLibPref Pref {
			get {
				NaGetLibPref prefCopied = new NaGetLibPref();
				
				if (pref == null) {
					LoadPref();
				}
				NaGet.Utils.FieldCopy(pref, ref prefCopied);
				
				return prefCopied;
			}
		}
		
		/// <summary>
		/// 設定ファイルのある場所を取得する
		/// </summary>
		public static string PrefPath {
			get {
				return Path.Combine(Environment.CurrentDirectory, "pref.xml");
			}
		}
		
		/// <summary>
		/// 設定ファイルを読み込む。または再読み込みする。
		/// </summary>
		public static void LoadPref()
		{
			string path = PrefPath;
			try {
				if (! File.Exists(path)) throw new FileNotFoundException(string.Empty, path);
				pref = NaGet.Utils.GetDeserializedObject<NaGetLibPref>(path);
			} catch(Exception) {
				pref = new NaGetLibPref();
			}
		}
		
		
		#endregion

		#region ファイルパス
		
		/// <summary>
		/// ファイルリスト提供レポジトリリストのファイル
		/// </summary>
		public static readonly string RepositoriesListFile = "repos.list.xml";

		/// <summary>
		/// ファイルリストから読み込まれたパッケージのリストファイル
		/// </summary>
		public static readonly string PackageListFile = "packages.list.xml";

		/// <summary>
		/// 本ソフトウェアを介してインストールされたパッケージのリストファイル
		/// </summary>
		public static readonly string ArchiveInstalledPackageListFile = "packages.envinstalled.xml";

		/// <summary>
		/// システムから検出されたパッケージのリストファイル
		/// </summary>
		public static readonly string SystemInstalledPackageListFile = "packages.sysinstalled.xml";
		
		/// <summary>
		/// インストールログファイル
		/// </summary>
		public static readonly string SystemInstalledPackageLogFile = "packages.sysinstalled.log.xml";
		
		/// <summary>
		/// アプリケーションデータを保存するフォルダーのパス
		/// </summary>
		public static string AppDataFolderPath
		{
			get {
				if (pref == null) LoadPref();
				
				if (! string.IsNullOrEmpty(pref.AppDataFolder)) {
					return pref.AppDataFolder;
				}
				
				return Environment.CurrentDirectory;
			}
		}

		/// <summary>
		/// インストーラーの一時置き場の親ディレクトリ
		/// </summary>
		public static string ArchiveFolderPath
		{
			get {
				if (pref == null) LoadPref();

				if (! string.IsNullOrEmpty(pref.CacheFolder)) {
					return pref.CacheFolder;
				}
				
				return Path.Combine(AppDataFolderPath, "Cache");
			
			}
		}
		
		/// <summary>
		/// アーカイバ方式のパッケージのインストール先フォルダー
		/// </summary>
		public static string ArchiveProgramFiles {
			get {
				return Path.Combine(AppDataFolderPath, "progs");
			}
		}
		
		/// <summary>
		/// アーカイバ方式のパッケージのプログラムグループフォルダー
		/// </summary>
		public static string ArchiveProgramGroup {
			get {
				return Path.Combine(AppDataFolderPath, "programs");
			}
		}

		/// <summary>
		/// アーカイバ方式のパッケージのSystem32のフォルダー
		/// </summary>
		public static string ArchiveSystem32 {
			get {
				return Path.Combine(ArchiveProgramFiles, ".system32");
			}
		}
		
		#endregion
		
		#region ネットワーク
		
		/// <summary>
		/// ウェブアクセスのためのプロキシを取得する。
		/// </summary>
		/// <returns>プロキシ</returns>
		public static IWebProxy WebProxy
		{
			get {
				if (pref == null) LoadPref();
				
				if (string.IsNullOrEmpty(pref.ProxyAddress)) {
					// 設定されていないときはシステムのデフォルトを使う
					IWebProxy proxy = WebRequest.GetSystemWebProxy();
					proxy.Credentials = CredentialCache.DefaultNetworkCredentials;
					return proxy;
				} else if (pref.ProxyAddress == "-") {
					// "-"のとき直接接続
					return null;
				} else {
					// host:portが設定されているならば、それをもとに設定
					WebProxy proxy = new WebProxy(pref.ProxyAddress);
					proxy.Credentials = CredentialCache.DefaultCredentials;
					
					// 認証プロキシの場合、認証情報を付加する
					try {
						Uri uri = new Uri(pref.ProxyAddress);
						if (! string.IsNullOrEmpty(uri.UserInfo)) {
							int pos = uri.UserInfo.IndexOf(':');
							if (pos >= 0) {
								proxy.Credentials = new NetworkCredential(
									uri.UserInfo.Substring(0, pos),
									uri.UserInfo.Substring(pos+1)
								);
							} else {
								proxy.Credentials = CredentialCache.DefaultCredentials;
							}
						}
					} catch (Exception) {
					}
					
					return proxy;
				}
			}
		}
		
		/// <summary>
		/// User-Agent文字列
		/// </summary>
		public static string UserAgentString {
			get {
				System.Reflection.Assembly asm = System.Reflection.Assembly.GetExecutingAssembly();
				string name = asm.GetName().Name;
				Version version = asm.GetName().Version;
				
				return string.Format("{0}/{1}.{2}.{3}", name, version.Major, version.Minor, version.Build);
			}
		}
		
		#endregion
		
		#region インストールプロセス
		
		/// <summary>
		/// インストーラーをウイルススキャンするか?
		/// </summary>
		public static bool EnableScanInstallerFile {
			get {
				if (pref == null) LoadPref();
				return pref.EnableScanInstallerFile;
			}
			set {
				if (pref == null) LoadPref();
				pref.EnableScanInstallerFile = value;
			}
		}

		/// <summary>
		/// インストールプロセスをバックグラウンドで行うか?
		/// </summary>
		public static bool InstallProcessOnBackground {
			get {
				if (pref == null) LoadPref();
				return pref.InstallOnBackground;
			}
			set {
				if (pref == null) LoadPref();
				pref.EnableScanInstallerFile = value;
			}
		}
		
		/// <summary>
		/// アップデートから除外するパッケージの名前
		/// </summary>
		public static string[] ExcludeUpdatePackageNames {
			get {
				if (pref == null) LoadPref();
				return pref.ExcludeUpdatePackageNames;
			}
			set {
				if (pref == null) LoadPref();
				pref.ExcludeUpdatePackageNames = value;
			}
		}
		
		#endregion
		
	}
}
